/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Save to raw, headerless iLBC data.
 *
 * Brian K. West <brian@bkw.org>
 * 
 * Copyright (C) 1999, Mark Spencer
 *
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */
 
#include <asterisk/lock.h>
#include <asterisk/channel.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/sched.h>
#include <asterisk/module.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#ifdef __linux__
#include <endian.h>
#else
#include <machine/endian.h>
#endif

struct frame {
    int stereo;
    int jsbound;
    int single;
    int lsf;
    int mpeg25;
    int header_change;
    int lay;
    int error_protection;
    int bitrate_index;
    int sampling_frequency;
    int padding;
    int extension;
    int mode;
    int mode_ext;
    int copyright;
    int original;
    int emphasis;
    int framesize; /* computed framesize */

    /* layer2 stuff */
    int II_sblimit;
    unsigned char *alloc;
};

struct ast_filestream {
	void *reserved[AST_RESERVED_POINTERS];
	/* This is what a filestream means to us */
	int fd; /* Descriptor */
	struct ast_frame fr;				/* Frame information */
	char waste[AST_FRIENDLY_OFFSET];		/* Buffer for sending frames, etc */
	char empty;					/* Empty character */
	struct frame mp3;				/* One Real MP3 Frame */
};


static ast_mutex_t mp3_lock = AST_MUTEX_INITIALIZER;
static int glistcnt = 0;

static char *name = "MP3";
static char *desc = "MP3 data format";
static char *exts = "mp3";

static long freqs[9] = { 44100, 48000, 32000,
			 22050, 24000, 16000 ,
			 11025 , 12000 , 8000 };

static int tabsel_123[2][3][16] = {
   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
     {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
     {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};


#define		FRAME_SYNC		0x7FF;
#define         MPG_MD_STEREO           0
#define         MPG_MD_JOINT_STEREO     1
#define         MPG_MD_DUAL_CHANNEL     2
#define         MPG_MD_MONO             3

struct id3tag {
	
        char header[10];
        char *data;
        int size;
        char flags;
        int hasfooter;
        int curpos;
};

static int decode_header_l(struct frame *fr,unsigned long newhead);
static int mp3_seek(struct ast_filestream *fs, long sample_offset, int whence);

/* decode_header is based on mpglib */
static int decode_header_l(struct frame *fr,unsigned long newhead)
{
    if( ((newhead>>21)&0x7ff) != 0x7ff) {
      fprintf(stderr, "Invalid Header (must start with 11 bits of 1s)\n");
      return(1);
    }
    if( newhead & (1<<20) ) {
      fr->lsf = (newhead & (1<<19)) ? 0x0 : 0x1;
      fr->mpeg25 = 0;
/*
    } else if (newhead & (1<<19)){
      	fprintf(stderr, "invalid type\n");
      return(1);
*/
    } else {
      fr->lsf = 1;
      fr->mpeg25 = 1;
    }

    if ((fr->lay = 4-((newhead>>17)&3)) == 4) {
      	fprintf(stderr, "Invalid layer\n");
	return(1);
    }

    if( ((newhead>>10)&0x3) == 0x3) {
      fprintf(stderr, "Invalid Header (bad sample rate)\n");
      return(1);
    }

    if(fr->mpeg25) {
      fr->sampling_frequency = 6 + ((newhead>>10)&0x3);
    }
    else
      fr->sampling_frequency = ((newhead>>10)&0x3) + (fr->lsf*3);
    fr->error_protection = ((newhead>>16)&0x1)^0x1;

    if(fr->mpeg25) /* allow Bitrate change for 2.5 ... */
      fr->bitrate_index = ((newhead>>12)&0xf);

    fr->bitrate_index = ((newhead>>12)&0xf);
    fr->padding   = ((newhead>>9)&0x1);
    fr->extension = ((newhead>>8)&0x1);
    fr->mode      = ((newhead>>6)&0x3);
    fr->mode_ext  = ((newhead>>4)&0x3);
    fr->copyright = ((newhead>>3)&0x1);
    fr->original  = ((newhead>>2)&0x1);
    fr->emphasis  = newhead & 0x3;

    fr->stereo    = (fr->mode == MPG_MD_MONO) ? 1 : 2;

    if(!fr->bitrate_index || fr->bitrate_index == 0xf)  // Don't support free format
    {
      fprintf(stderr, "Invalid Header (bad bitrate) %d\n", fr->bitrate_index);
      return (1);
    }

    switch(fr->lay)
    {
      case 1:
	fr->framesize  = (long) tabsel_123[fr->lsf][0][fr->bitrate_index] * 12000;
	fr->framesize /= freqs[fr->sampling_frequency];
	fr->framesize  = ((fr->framesize+fr->padding)<<2);
        break;
      case 2:
	fr->framesize = (long) tabsel_123[fr->lsf][1][fr->bitrate_index] * 144000;
	fr->framesize /= freqs[fr->sampling_frequency];
	fr->framesize += fr->padding;
        break;
      case 3:
	fr->framesize  = (long) tabsel_123[fr->lsf][2][fr->bitrate_index] * 144000;
	fr->framesize /= freqs[fr->sampling_frequency]<<(fr->lsf);
	fr->framesize = fr->framesize + fr->padding;
        break;
      default:
        fprintf(stderr,"Sorry, unknown layer type.\n");
        return (1);
    }
    return 0;
}


static struct ast_filestream *mp3_open(int fd)
{
	/* We don't have any header to read or anything really, but
	   if we did, it would go here.  We also might want to check
	   and be sure it's a valid file.  */
	struct ast_filestream *tmp;
	if ((tmp = malloc(sizeof(struct ast_filestream)))) {
		memset(tmp, 0, sizeof(struct ast_filestream));
		tmp->fd = fd;
		tmp->fr.data = tmp->mp3.alloc;
		tmp->fr.frametype = AST_FRAME_VOICE;
		tmp->fr.subclass = AST_FORMAT_MP3;
		/* datalen will vary for each frame */
		tmp->fr.src = name;
		tmp->fr.mallocd = 0;
		tmp->fr.datalen = 417;  /* Default to usual size of a 128kbit file */
		tmp->fr.data = (char *)malloc(tmp->fr.datalen * sizeof(char));
		if (ast_mutex_lock(&mp3_lock)) {
			ast_log(LOG_WARNING, "Unable to lock mp3 list\n");
			free(tmp);
			return NULL;
		}
		glistcnt++;
		ast_mutex_unlock(&mp3_lock);
		ast_update_use_count();
	}
	return tmp;
}

static struct ast_filestream *mp3_rewrite(int fd, char *comment)
{
	/* We don't have any header to read or anything really, but
	   if we did, it would go here.  We also might want to check
	   and be sure it's a valid file.  */
	struct ast_filestream *tmp;
	if ((tmp = malloc(sizeof(struct ast_filestream)))) {
		memset(tmp, 0, sizeof(struct ast_filestream));
		tmp->fd = fd;
		if (ast_mutex_lock(&mp3_lock)) {
			ast_log(LOG_WARNING, "Unable to lock mp3 list\n");
			free(tmp);
			return NULL;
		}
		glistcnt++;
		ast_mutex_unlock(&mp3_lock);
		ast_update_use_count();
	} else
		ast_log(LOG_WARNING, "Out of memory\n");
	return tmp;
}

static void mp3_close(struct ast_filestream *s)
{
	if (ast_mutex_lock(&mp3_lock)) {
		ast_log(LOG_WARNING, "Unable to lock mp3 list\n");
		return;
	}
	glistcnt--;
	ast_mutex_unlock(&mp3_lock);
	ast_update_use_count();
	close(s->fd);
	free(s->mp3.alloc);
	free(s);
	s = NULL;
}

static struct ast_frame *mp3_read(struct ast_filestream *s, int *whennext)
{
	int res=0;
	unsigned long head=0;
	unsigned char hdr[4];
	/* Send a frame from the file to the appropriate channel */
	s->fr.frametype = AST_FRAME_VOICE;
	s->fr.subclass = AST_FORMAT_MP3;
	s->fr.offset = AST_FRIENDLY_OFFSET;
	s->fr.samples = 208;
	s->fr.mallocd = 0;
	while (1) {
		if ((res = read(s->fd, hdr, 1)) != 1) {
			return NULL;
		}
		if (hdr[0] ^ 0xff)
			continue;

		if ((res = read(s->fd, hdr+1, 3)) != 3)
			return NULL;

		head = hdr[0] & 0xff;
		head <<= 8;
		head |= hdr[1] & 0xff;
		head <<= 8;
		head |= hdr[2] & 0xff;
		head <<= 8;
		head |= hdr[3] & 0xff;

		if(!(res = decode_header_l(&s->mp3,head))) {
			if(s->fr.datalen != s->mp3.framesize) {
				if (s->mp3.framesize > 156)  /* That's too small.... */
					goto next;
				free(s->fr.data);
				s->fr.data = (char *)malloc(s->mp3.framesize * sizeof(char));
				s->fr.datalen = s->mp3.framesize;
			}
			memcpy(s->fr.data, hdr, 4);
			if((res = read(s->fd, s->fr.data+4, s->mp3.framesize-4)) == s->mp3.framesize-4) {
				break;
			} else {
				ast_log(LOG_WARNING, "short read %d/%d!\n", res+4, s->mp3.framesize);
				return NULL;
			}
		}
next:
		mp3_seek(s, -3, SEEK_CUR);
	}
	/* If we get here, we have a valid frame in memory */
	*whennext = s->fr.samples;
	return &s->fr;
}

static int mp3_write(struct ast_filestream *fs, struct ast_frame *f)
{
	int res;
	if (f->frametype != AST_FRAME_VOICE) {
		ast_log(LOG_WARNING, "Asked to write non-voice frame!\n");
		return -1;
	}
	if (f->subclass != AST_FORMAT_MP3) {
		ast_log(LOG_WARNING, "Asked to write non-MP3 frame (%d)!\n", f->subclass);
		return -1;
	}
	/*
	if (f->datalen % 50) {
		ast_log(LOG_WARNING, "Invalid data length, %d, should be multiple of 50\n", f->datalen);
		return -1;
	}
	*/
	if ((res = write(fs->fd, f->data, f->datalen)) != f->datalen) {
			ast_log(LOG_WARNING, "Bad write (%d): %s\n", res, strerror(errno));
			return -1;
	}
	return 0;
}

static char *mp3_getcomment(struct ast_filestream *s)
{
	return NULL;
}

static int mp3_seek(struct ast_filestream *fs, long sample_offset, int whence)
{
	long bytes;
	off_t min,cur,max,offset=0;
	min = 0;
	cur = lseek(fs->fd, 0, SEEK_CUR);
	max = lseek(fs->fd, 0, SEEK_END);
	
	bytes = sample_offset;
	if (whence == SEEK_SET)
		offset = bytes;
	else if (whence == SEEK_CUR || whence == SEEK_FORCECUR)
		offset = cur + bytes;
	else if (whence == SEEK_END)
		offset = max - bytes;
	if (whence != SEEK_FORCECUR) {
		offset = (offset > max)?max:offset;
	}
	// protect against seeking beyond begining.
	offset = (offset < min)?min:offset;
	if (lseek(fs->fd, offset, SEEK_SET) < 0)
		return -1;
	return 0;
}

static int mp3_trunc(struct ast_filestream *fs)
{
	/* Truncate file to current length */
	if (ftruncate(fs->fd, lseek(fs->fd, 0, SEEK_CUR)) < 0)
		return -1;
	return 0;
}

static long mp3_tell(struct ast_filestream *fs)
{
	off_t offset;
	offset = lseek(fs->fd, 0, SEEK_CUR);
	return offset;
}

int load_module()
{
	return ast_format_register(name, exts, AST_FORMAT_MP3,
								mp3_open,
								mp3_rewrite,
								mp3_write,
								mp3_seek,
								mp3_trunc,
								mp3_tell,
								mp3_read,
								mp3_close,
								mp3_getcomment);
								
								
}

int unload_module()
{
	return ast_format_unregister(name);
}	

int usecount()
{
	int res;
	if (ast_mutex_lock(&mp3_lock)) {
		ast_log(LOG_WARNING, "Unable to lock mp3 list\n");
		return -1;
	}
	res = glistcnt;
	ast_mutex_unlock(&mp3_lock);
	return res;
}

char *description()
{
	return desc;
}


char *key()
{
	return ASTERISK_GPL_KEY;
}
